/* Democracy - a Bukkit plugin bringing voting features to Minecraft Servers
 * Copyright (c) 2011 Ed Halley <ed@halley.cc>
 * Offered under the MIT License.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package cc.halley.bukkit.Democracy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import org.bukkit.ChatColor;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.util.config.Configuration;

public class DemocracyProposeCommand
    implements CommandExecutor
{
	protected Democracy democracy;

	public DemocracyProposeCommand(Democracy democracy)
		{ this.democracy = democracy; }

	public boolean onHelpPropose(CommandSender sender)
	{
		sender.sendMessage(ChatColor.GRAY + "usage: /propose [page]: show pending issues");
		sender.sendMessage(ChatColor.GRAY + "usage: /propose <issue> [-options] [text]");
		sender.sendMessage(ChatColor.GRAY + "  -cancel:         cancels the proposal");
		sender.sendMessage(ChatColor.GRAY + "  -rollcall:       voters vote /aye to join");
		sender.sendMessage(ChatColor.GRAY + "  -choices x,y,z:  voters /vote x or y or z");
		sender.sendMessage(ChatColor.GRAY + "  -duration 2h:    proposal open for 2 hours");
		sender.sendMessage(ChatColor.GRAY + "  -open:     opens the proposal for votes");
		sender.sendMessage(ChatColor.GRAY + "Proposal can be altered options until -open.");
		return true;
	}

	public boolean onHelpNominate(CommandSender sender)
	{
		sender.sendMessage(ChatColor.GRAY + "usage: /nominate <player> [-options] [text]");
		sender.sendMessage(ChatColor.GRAY + "  -cancel:         cancels the nomination");
		sender.sendMessage(ChatColor.GRAY + "  -duration 2h:    nomination open for 2 hours");
		sender.sendMessage(ChatColor.GRAY + "  -open:           opens the nomination for votes");
		sender.sendMessage(ChatColor.GRAY + "Nomination can be altered options until -open.");
		sender.sendMessage(ChatColor.GRAY + "You cannot nominate yourself.  Making a");
		sender.sendMessage(ChatColor.GRAY + "nomination assumes you vote /aye for yourself.");
		return true;
	}

	public boolean listMyProposals(CommandSender sender, int page)
	{
		String who = sender.getName();
		Configuration config = democracy.getConfiguration();
		boolean accountable = config.getBoolean("voting.accountable-vetoes",
		                                        true);
		List<DemocracyIssue> issues = new ArrayList<DemocracyIssue>();
		Iterator<DemocracyIssue> it = democracy.getAllIssues();
		while (it.hasNext())
		{
			DemocracyIssue issue = it.next();
			if (!issue.proponent.equals(who))
				continue;
			issues.add(issue);
		}

		int PAGE = 5;
		
		page--;
		if (page*PAGE >= issues.size())
		{
			sender.sendMessage(ChatColor.GRAY +
			                   "You have no more proposals to review.");
			return false;
		}
		
		Collections.sort(issues, new Comparator<DemocracyIssue>() {
            public int compare(DemocracyIssue a, DemocracyIssue b)
            	{ return a.name.compareTo(b.name); } } );

		sender.sendMessage(ChatColor.GRAY +
		                   "Your pending proposals, " +
		                   ChatColor.GOLD +
		                   "page " + (page+1) +
		                   " of " + ((issues.size()+PAGE-1)/PAGE));
		for (int i = page*PAGE; i < page*PAGE+PAGE; i++)
		{
			if (i >= issues.size())
				break;
			DemocracyIssue issue = issues.get(i);
			sender.sendMessage(ChatColor.GRAY +
			                   issue.summarizeStatus(who, accountable));
		}

		return true;
	}

    @Override
	public boolean onCommand(CommandSender sender, Command command,
	                         String label, String[] args)
	{
    	democracy.checkIssuesDue();
    	
		// No issue to propose?
		int page = 0;
		try { page = (args.length == 0)? 1 : Integer.parseInt(args[0]); }
		catch (Exception e) { ; }
		if (page > 0)
		{
			listMyProposals(sender, page);
			return true;
		}

		// name or subcommand first
		int argp = 0;
		String name = args[argp++];
		if (name.startsWith("-"))
		{
			sender.sendMessage(ChatColor.RED +
			                   "Proposals must be given a name first.");
			return false;
		}
		
		if (name.equalsIgnoreCase("help"))
			return "nominate".equalsIgnoreCase(label)?
                onHelpNominate(sender) : onHelpPropose(sender);

		if (DemocracyIssue.isIssueNameReserved(name))
		{
			sender.sendMessage(ChatColor.RED +
			                   "Cannot use '" + name + "' as a proposal name.");
			return false;
		}

		// must be a name
		String who = sender.getName();
		DemocracyIssue issue = democracy.findIssue(name);
		Configuration config = democracy.getConfiguration();
		boolean addIssue = false;
		
		// if name was not an existing issue, make one
		if (issue == null)
		{
			if (tooManyProposals(config, sender))
			{
				sender.sendMessage(ChatColor.RED + 
				                   "Wait for some proposals to clear out.");
				return true;
			}
			
			issue = new DemocracyIssue(name, who);
			addIssue = true;
			
			if ("nominate".equalsIgnoreCase(label))
				issue.type = DemocracyIssue.NOMINATION;
			
			// set options to plugin defaults
			issue.minimum = config.getInt("default.minimum", 5);
			issue.secret = config.getBoolean("default.secrecy", false);
			issue.expires = parseDuration(config.getString("default.duration"));
			if (issue.expires < 0)
				issue.expires = 5*60*60*1000;
		}
		else
		{
			if (!who.equals(issue.proponent))
			{
				sender.sendMessage(ChatColor.RED +
				                   issue.proponent +
				                   " already made that proposal.");
				return false;
			}
			if (issue.isClosed())
			{
				sender.sendMessage(ChatColor.RED + 
				                   "That proposal has already been resolved.");
				return false;
			}
			if (issue.isOpened())
			{
				sender.sendMessage(ChatColor.RED + 
				                   "That proposal is already open for voting.");
				return false;
			}
		}

		boolean summarizeOptions = false;
		boolean openIssue = false;
		boolean castAye = false;

		// then options
		while (argp < args.length && args[argp].startsWith("-"))
		{
			String option = args[argp++].substring(1);
			while (option.startsWith("-"))
				option = option.substring(1);

			// -cancel
			if (option.equalsIgnoreCase("cancel") ||
				option.equalsIgnoreCase("close") ||
				option.equalsIgnoreCase("end") ||
				option.equalsIgnoreCase("abort"))
			{
				democracy.removeIssue(issue);
				sender.sendMessage(ChatColor.GREEN +
				                   "The proposal has been canceled.");
				return true;
			}

			// -open
			if (option.equalsIgnoreCase("open") ||
				option.equalsIgnoreCase("ready") ||
				option.equalsIgnoreCase("begin") ||
				option.equalsIgnoreCase("start"))
			{
				openIssue = true;
				if ("propose".equalsIgnoreCase(label) &&
					Democracy.config.getBoolean(
					    "voting.proposing-includes-aye-vote", false))
					castAye = true;
				else if ("nominate".equalsIgnoreCase(label) &&
						Democracy.config.getBoolean(
						    "voting.nominating-includes-aye-vote", false))
						castAye = true;
			}

			// -rollcall
			if (option.equalsIgnoreCase("rollcall") ||
				option.equalsIgnoreCase("roll"))
			{
				issue.type = DemocracyIssue.ROLLCALL;
				issue.setChoices(null);
				summarizeOptions = true;
			}

			// -aye or -ayenay
			if (option.equalsIgnoreCase("aye") ||
				option.equalsIgnoreCase("nay") ||
				option.equalsIgnoreCase("aye,nay") ||
				option.equalsIgnoreCase("up,down") ||
				option.equalsIgnoreCase("yes,no"))
			{
				if (issue.type == DemocracyIssue.CHOICE)
					issue.type = DemocracyIssue.AYE;
				issue.setChoices(null);
				summarizeOptions = true;
			}

			// -choices [x,y,z]
			// -aye or -ayenay
			if (option.equalsIgnoreCase("choice") ||
				option.equalsIgnoreCase("choices"))
			{
				if (issue.type == DemocracyIssue.NOMINATION)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Nominations must be aye/nay choices.");
					return true;
				}
				if (issue.type == DemocracyIssue.ROLLCALL)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Roll-call votes have aye/nay choices.");
					return true;
				}
				if (argp >= args.length)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Provide a comma-separated list, " +
					                   "such as /propose flavor " +
					                   "-choices vanilla,chocolate,banana");
					return true;
				}
					
				issue.type = DemocracyIssue.CHOICE;
				issue.setChoices(args[argp++].split(","));
				//TODO: check for too few choices
				summarizeOptions = true;
			}

			// -players
			if (option.equalsIgnoreCase("players"))
			{
				// choices are freeform, finding best fit player name
				// for "who do you think is most bad-ass?"
			}

			// -secret or -public
			if (option.equalsIgnoreCase("secret"))
			{
				if (config.getBoolean("voting.allow-secret-vote", true))
				{
					issue.secret = true;
					sender.sendMessage(ChatColor.GRAY +
					                   "Votes will not be announced.");
				}
				else
					sender.sendMessage(ChatColor.GRAY +
					                   "Votes cannot be kept secret.");
			}
			if (option.equalsIgnoreCase("public"))
			{
				if (config.getBoolean("voting.allow-public-vote", true))
				{
					issue.secret = false;
					sender.sendMessage(ChatColor.GRAY +
					                   "Votes will be announced.");
				}
				else
					sender.sendMessage(ChatColor.GRAY +
					                   "Votes will not be announced.");
			}

			// -minimum [count]
			if (option.equalsIgnoreCase("minimum"))
			{
				if (argp >= args.length)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Provide a minimum number of votes, " +
					                   "such as /propose spleef " +
					                   "-minimum 6 to get at least six " +
					                   "aye or nay votes total.");
					return true;
				}
				int minimum = -1;
				try { minimum = Integer.parseInt(args[argp++]); }
				catch (Exception e) { ; }
				if (minimum < 1)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Minimum was not a number.");
					return true;
				}
				
				sender.sendMessage(ChatColor.GRAY +
				                   "Proposal will need " + minimum +
				                   " votes (of any type) " +
				                   "or it expires without an answer.");
				issue.minimum = minimum;
			}

			// -duration [count][units]
			if (option.equalsIgnoreCase("duration"))
			{
				if (argp >= args.length)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Provide a time span for votes, " +
					                   "such as /propose pillowfight " +
					                   "-duration 5m for a five minute vote.");
					return true;
				}
				long expires = parseDuration(args[argp++]);
				if (expires < 0)
				{
					sender.sendMessage(ChatColor.RED +
					                   "Duration was not understood. " +
					                   "Examples:  120 (seconds) " +
					                   "or 45m (minutes) " +
					                   "or 2h (hours) or 4d (days).");
					return true;
				}
				
				issue.expires = expires;
				sender.sendMessage(ChatColor.GRAY +
				                   "Proposal will allow votes for " +
				                   issue.summarizeDuration(expires) + ".");
			}
			
		}

		// then description
		if (argp < args.length)
		{
    		issue.description = "";
    		StringBuilder sb = new StringBuilder();
    		while (argp < args.length)
    		{
    			sb.append(issue.description); issue.description = " ";
    			sb.append(args[argp++]);
    		}
    		issue.description = sb.toString();
		}

		// was this brand-new or just reconfiguring?
		if (addIssue)
		{
			democracy.addIssue(issue);
			sender.sendMessage(ChatColor.GREEN +
							   "You've proposed a new issue to vote.");
		}
		
		// did they include the -open?
		if (openIssue)
		{
			issue.open();
			democracy.announceIssue(issue);
			if (castAye && sender instanceof Player)
			{
				boolean secret = issue.secret;
				issue.secret = true;
				issue.castVote((Player)sender, "aye");
				issue.secret = secret;
			}
		}
		else
		{
			if (summarizeOptions)
				sender.sendMessage(issue.summarizeOptions());
    		sender.sendMessage(ChatColor.GRAY +
			                   "To open for voting: " +
    		                   "/" + label + " " + name + " -open");
		}

		return true;
	}

	private boolean tooManyProposals(Configuration config, CommandSender who)
    {
		int limit = config.getInt("limits.open-proposals", 5);
	    if (democracy.countAllIssues() >= limit)
	    	return true;
	    if (who instanceof Player)
	    {
	    	Player player = (Player)who;
    		limit = config.getInt("limits.open-proposals-per-player", 1);
    	    if (democracy.countProposedIssues(player) >= limit)
    	    	return true;
	    }
		return false;
    }

	private long parseDuration(String string)
    {
		string = string.toLowerCase().trim();
		int seconds = 0;
		int unit = 1;
		if (string.endsWith("s"))
			{ unit = 1; string = string.substring(0, string.length()-1); }
		else if (string.endsWith("m"))
			{ unit = 60; string = string.substring(0, string.length()-1); }
		else if (string.endsWith("h"))
			{ unit = 60*60; string = string.substring(0, string.length()-1); }
		else if (string.endsWith("d"))
			{ unit = 60*60*24; string = string.substring(0, string.length()-1); }
		string = string.trim();
		try
			{ seconds = Integer.parseInt(string); }
		catch (Exception e)
			{ return -1; }
		if (seconds < 1)
			return -1;
	    return seconds * unit * 1000L;
    }
}
